// Authors: Sven Förster, Duc Thanh Tran

#ifndef EQUITABLECOLORING_HPP
#define EQUITABLECOLORING_HPP

#include "boost.hpp"
#include "input.hpp"
#include "structs.hpp"
#include "status.hpp"
#include "timer.hpp"

#include <boost/dynamic_bitset.hpp>

#include <memory>

namespace EqDsatur {

using Clique = std::vector<Vertex>;
using IndCliqueDecomposition = std::vector<std::vector<Vertex>>;

/// Computes an equitable coloring for an undirected graph
class EquitableColoring {
public:
	EquitableColoring() = delete;
	explicit EquitableColoring(const Parameters &pArg);
	~EquitableColoring() = default;

private:
	Graph graph_;
	std::vector<std::vector<int>> forbiddenColorMap_;
	std::vector<int> cliqueMap_;
	std::vector<int> colorMap_;
	std::vector<int> colorClassCardinality_;
	std::vector<int> saturationDegreeMap_;
	std::vector<int> rankMap_;
	std::vector<int> cliqueIndex_;

	Parameters parameters_;
	FreeColors freeColors_;
	IndCliqueDecomposition indClq_;
	Cliques cliques_;
	Bounds bounds_;
	CurrentStatus currentStatus_;
	Counts counts_;
	Backtracking backtrack_;
	Times times_;
	std::unique_ptr<Timer> startTimer_;

	std::vector<int> degreeMap_;

	inline void setBounds(const int LB, const int UB) noexcept {
		bounds_.LB = LB;
		bounds_.UB = UB;
	}

	inline void resetCliqueIndices() noexcept {
		std::fill(cliqueIndex_.begin(), cliqueIndex_.end(), Status::to_underlying(CliqueIndex::RestVertex));
	}

	inline bool isEquitableColoring() const noexcept {
		return isColoring() && isEquitable();
	}

	inline bool isUncoloredNotInClique(const Vertex vertex) const noexcept {
		return isUncolored(vertex) && cliqueMap_[static_cast<int>(vertex)] == Status::to_underlying(CliqueStatus::NoClique);
	}

	inline bool isUncolored(const Vertex vertex) const noexcept {
		return colorMap_[static_cast<int>(vertex)] == Status::to_underlying(VertexColor::Uncolored);
	}

	inline bool isColorAvailable(const Vertex vertex, const int color) const noexcept {
		return forbiddenColorMap_[static_cast<int>(vertex)][color-1] == Status::to_underlying(ForbiddenColorStatus::Available);
	}

	inline long getCurrentUpperBound() const noexcept {
		return std::min(bounds_.UB-1, currentStatus_.nColors+1);
	}

	inline int getColorClassSize(const int color) const noexcept {
		return colorClassCardinality_[color-1];
	}

	inline bool isRestVertex(const Vertex vertex) const noexcept {
		return cliqueIndex_[vertex] == Status::to_underlying(CliqueIndex::RestVertex);
	}

	inline int getDegree(const Vertex vertex) const noexcept {
		return degreeMap_[static_cast<int>(vertex)];
	}

	inline bool willBeFreeColor(const Vertex neighbor, const int color) const noexcept {
		return forbiddenColorMap_[neighbor][color-1] == 1;
	}

	inline int getCliqueIndex(const Vertex vertex) const noexcept {
		return cliqueIndex_[vertex];
	}

	inline int getExclusiveColor(const Vertex vertex) const noexcept {
		return freeColors_.exclusiveColor[vertex]-1;
	}

	inline bool hasExclusiveColor(const Vertex vertex) const noexcept {
		return freeColors_.exclusiveColor[vertex] != Status::to_underlying(ExclusiveColor::None);
	}

	inline void resetExclusiveColorToNone(const Vertex vertex) noexcept {
		freeColors_.exclusiveColor[vertex] = Status::to_underlying(ExclusiveColor::None);
	}

	/// Vertex selection strategy
	Vertex passVSS() noexcept;

	/// Uses saturation degree in order to break ties during vertex selection
	int pass(const Vertex vertex, const int maxSatDeg) const noexcept;

	/// Computes the maximum saturation degree with current coloring
	int findMaxSatDeg() const noexcept;
	std::vector<Vertex> findVertexSatDeg(const int satDeg) const noexcept;

	bool checkForBacktracking(const Vertex vertex) noexcept;
	void newUBBacktracking();

	IndCliqueDecomposition computeIndependentCliques() noexcept;
	Clique findMaxClique() const noexcept;
	Clique findMaxCliqueWithVertex(const Vertex vertex, const int maxCliqueSize) const noexcept;

	/// tries to find a new (better) independent clique decomposition after coloring a vertex
	void updateIndepCliques(const Vertex vertex) noexcept;
	bool foundBetterIndCliqueDecomposition() noexcept;
	/// not used; may be interesting for future implementations/uses
	int countFreeColors(const IndCliqueDecomposition &cliqueDecomposition) const noexcept;

	/// Used to break ties when computing a maximum clique
	int calculateDegreeSum(const Clique &clique) const noexcept;
	void removeVertexFromIndependentCliques(const Vertex vertex) noexcept;

	void colorStartClique(const Clique &clique) noexcept;
	void colorVertex(const Vertex vertex, const int color) noexcept;
	void uncolorVertex(const Vertex vertex) noexcept;

	void printOutput() const noexcept;

	void updateFreeColorsNewCliqueDecomposition() noexcept;
	void initialization() noexcept;
	void resizeMapsAndIndexes() noexcept;

	/// Updates clique information by assigning an index
	/// to our computed clique among others.
	bool putInClique(const Clique &clique) noexcept;

	/// Sets information needed for backtracking
	int backtrackToRank(const Vertex v) const noexcept;

	/// FBC stands for forbiddenColorMap. When we color a vertex v with c, we disallow c
	/// for neighboring vertices of v.
	void addFBC(const Vertex vertex, const int color) noexcept;
	void removeFBC(const Vertex vertex, const int color) noexcept;
	void incColorClass(const int color) noexcept;
	void decColorClass(const int color) noexcept;

	/// Updates size information regarding the biggest color class (M) and the amount
	/// of color classes of size M (T).
	void updateTandM(const int lastUsedColor, const bool wasIncremented) noexcept;

	bool checkOvertime() const noexcept;
	void setClique(const int nodesInCliques, const int numberCliques) noexcept;

	/// -------------
	/// Free colors update methods: we store information about free color counts during execution (if enabled)
	/// and thus enable three combinatorial pruning rules
	/// -------------
	void initFreeColors() noexcept;
	void setCliqueIndices(const Clique &maxClique, const int cliqueIndex) noexcept;
	bool needUpdateCombPruning() const noexcept;
	bool needUpdateExclusiveColor() const noexcept;

	void moveToRestVertex(const Vertex vertex) noexcept;
	void updateCliqueDecompCliqueRemoval(const int removedCliqueIndex) noexcept;
	void resetCountersFreeColors() noexcept;
	void resizeCliqueFreeColors() noexcept;
	void updateVertexFreeColors(const Vertex vertex) noexcept;

	void colorVertexFreeColors(const Vertex vertex, const int color) noexcept;
	void updateNeighborsAfterColoring(const Vertex vertex, const int color) noexcept;
	void updateNeighborVertexColoring(const Vertex vertex, const int color) noexcept;
	void checkForExclusiveColorColoring(const Vertex vertex, const int bannedColor) noexcept;
	void disableFreeColors(const Vertex vertex) noexcept;

	void uncolorVertexFreeColors(const Vertex vertex, const int color) noexcept;
	void updateNeighborsAfterUncoloring(const Vertex vertex, const int color) noexcept;
	void updateNeighborVertexUncoloring(const Vertex vertex, const int color) noexcept;

	void checkForExclusiveColorUncoloring(const Vertex vertex, const int newColor) noexcept;
	void enableFreeColors(const Vertex vertex) noexcept;

	bool isPrunableCombPruning_1(const int colorAmount) noexcept;
	bool isPrunableCombPruning_2(const int colorAmount) noexcept;
	bool isPrunableCombPruning_3() noexcept;

	std::tuple<int, int, int> computeExclusiveColors(const int selectedAndBannedColor, const int colorAmount) const noexcept;
	std::tuple<int, int, int> computeCliqueVerticesWithFreeColors(const int selectedAndBannedColor, const int colorAmount) const noexcept;
	std::tuple<int, int, int> computeRestVerticesWithFreeColors(const int selectedAndBannedColor, const int colorAmount) const noexcept;
	std::tuple<int, int, int> computeMinCapacities(const int selectedAndBannedColor, const int colorAmount) const noexcept;
	std::tuple<int, int, int> computeMaxCapacities(const int selectedAndBannedColor, const int colorAmount) const noexcept;

	/// -------------
	/// Initial bound related
	/// -------------

	/// Computes a maximum clique and uses its size of a lower bound.
	/// Furthermore we color this clique.
	void calculateLowerBoundAndColorStartClique() noexcept;

	/// We compute our upper bound by coloring greedily and afterwards
	/// swap colors in order to get an equitable coloring.
	void calculateUpperBound() noexcept;

	int naiveUB() noexcept;
	void greedyColoring() noexcept;
	bool swapNodeToColor(const Vertex vertex, const int color) noexcept;

	bool isEquitable() const noexcept;
	bool isColoring() const noexcept;

	/// -------------
	/// Maximum Flow Pruning related
	/// -------------

	MaxFlow maxFlow_;
	PropertyMapsMaxFlow propertyMapsMaxFlow_;

	void initPropertyMapsMaxFlow() noexcept;

	/// Mendez-Diaz et. al D-Satur algorithm
	bool dSatur() noexcept;

	/// Our algorithm which solves maximum flow problems in order to prune branches. Furthermore,
	/// combinatorial pruning rules can be enabled as well.
	bool dSaturClique() noexcept;

	bool isPrunable(const Vertex vertex) noexcept;
	bool isPrunableColor(const int colorAmount) const noexcept; // TODO: delete; not used
	bool isPrunableTheoremOne() const noexcept;


	bool isPrunableMaxFlow(const int colors) noexcept;
	long solveMaxFlow(const VertexFord source, const VertexFord target) noexcept;

	/// Initialized our flow-graph
	void initBackupGraphs(const int colors) noexcept;

	/// Builds arcs needed for flow-graph and maximum flow problem
	void initNetwork(const int colors) noexcept;
	/// rearranges arcs and capacities in order to solve the second maximum flow problem
	void initNetwork2(const int colors) noexcept;

	void addEdgesIntoFlowNetwork(const VertexFord source, const VertexFord target, long capacityFwd, long capacityBwd) noexcept;
	void modifyEdgesFlowNetwork(const VertexFord source, const VertexFord target, long capacityFwd, long capacityBwd) noexcept;

	/// not used
	void printResidualCap() const noexcept;
	/// not used
	void resetCapacities() const noexcept;
	/// not used
	void resetResidualCapacity() noexcept;
};

} // namespace EqDsatur

#endif
